from .base import SearchHelperMixin
from ...searchset.mapping import ElasticMapping


class ElasticSearchHelperMixin(SearchHelperMixin):
    _mapping_class = ElasticMapping

    @classmethod
    def _format_mapping_values(cls, mapping, prefix="") -> list:
        keys = []

        for key, value in mapping.items():
            keys.append(prefix + key)
            if "properties" in value:
                keys.extend(cls._format_mapping_values(value["properties"], "".join([prefix, key, "."])))

        return keys

    @classmethod
    def _get_raw_mapping(cls) -> list:
        model_instance = cls.Meta.model()
        mapping = cls.Meta.model._get_es_client().indices.get_mapping(index=model_instance._get_index())
        if not mapping:
            return []
        last_index = max(mapping.keys())

        try:
            mapping = mapping[last_index]['mappings'][model_instance._doc_type.name]['properties']
        except KeyError:
            return []

        keys = cls._format_mapping_values(mapping)
        return keys



from django.core.exceptions import FieldError

from .base import BaseSearchSet
from ..searchset.fields.django import DjangoSearchField, DjangoSearchFieldWithoutWildcard
from ..parser import LuceneToDjangoParserMixin


class DjangoSearchSet(LuceneToDjangoParserMixin, BaseSearchSet):
    _field_base_class = DjangoSearchFieldWithoutWildcard
    _default_field = DjangoSearchField

    @classmethod
    def filter(cls, queryset, search_terms):
        query = cls.parse(raw_expression=search_terms)
        try:
            return queryset.filter(query)
        except FieldError:
            return queryset.none()



import warnings

warnings.warn(
    "Import from base dir will be deprecated in version 0.2.0, use lucyfer.searchset.fields.base import instead",
    DeprecationWarning
)

from ..searchset.fields.base import *



import warnings

warnings.warn(
    "Import from base dir will be deprecated in version 0.2.0, use lucyfer.parser.base import instead",
    DeprecationWarning
)

from ..parser.base import BaseLuceneParserMixin



import warnings

warnings.warn(
    "Import from elastic dir will be deprecated in version 0.2.0, use lucyfer.backend import instead",
    DeprecationWarning
)

from ..backend import ElasticLuceneSearchFilterMixin



import warnings

warnings.warn(
    "Import from elastic dir will be deprecated in version 0.2.0, use lucyfer.searchset.fields import instead",
    DeprecationWarning
)

from ..searchset.fields.elastic import *



import warnings

from django.db.models import Q
from lucyparser.tree import Operator

from .base import BaseSearchField, negate_query_if_necessary
from ...utils import LuceneSearchCastValueException


class DjangoSearchFieldWithoutWildcard(BaseSearchField):
    DEFAULT_LOOKUP = "icontains"

    def create_query_for_sources(self, condition):
        query = Q()

        lookup = self.get_lookup(condition.operator)
        value = self.cast_value(condition.value)

        for source in self.get_sources(condition.name):
            query = query | Q(**{"{}__{}".format(source, lookup): value})
        return query

    @negate_query_if_necessary
    def get_query(self, condition):
        if self.match_all(value=condition.value):
            return None

        return self.create_query_for_sources(condition=condition)


class DjangoSearchField(DjangoSearchFieldWithoutWildcard):
    def create_query_for_sources(self, condition):
        value = self.cast_value(condition.value)

        if value.startswith("*") and value.endswith("*"):
            lookup = "icontains"
            value = value[1:-1]
        elif value.startswith("*"):
            lookup = "iendswith"
            value = value[1:]
        elif value.endswith("*"):
            lookup = "istartswith"
        else:
            return super().create_query_for_sources(condition)

        if not value:
            return None

        sources = self.get_sources(condition.name)

        query = Q()

        for source in sources:
            query = query | Q(**{"{}__{}".format(source, lookup): value})

        return query


class DjangoCharField(DjangoSearchField):
    OPERATOR_TO_LOOKUP = {
        Operator.EQ: "icontains",
        Operator.NEQ: "iexact",
    }


class DjangoNumberField(DjangoSearchFieldWithoutWildcard):
    OPERATOR_TO_LOOKUP = {
        Operator.GTE: "gte",
        Operator.LTE: "lte",
        Operator.GT: "gt",
        Operator.LT: "lt",
        Operator.EQ: "exact",
        Operator.NEQ: "exact",
    }


class DjangoIntegerField(DjangoNumberField):
    def cast_value(self, value):
        try:
            return int(value)
        except (ValueError, TypeError):
            raise LuceneSearchCastValueException()


class DjangoFloatField(DjangoNumberField):
    def cast_value(self, value):
        try:
            return float(value)
        except (ValueError, TypeError):
            raise LuceneSearchCastValueException()


class DjangoBooleanField(DjangoSearchFieldWithoutWildcard):
    OPERATOR_TO_LOOKUP = {
        Operator.EQ: "exact",
        Operator.NEQ: "exact",
    }

    _values = {"true": True, "false": False}
    _default_get_available_values_method = _values.keys

    def cast_value(self, value):
        value = value.lower()
        if value in self._values:
            return self._values[value]

        raise LuceneSearchCastValueException()


class DjangoNullBooleanField(DjangoBooleanField):
    _values = {"true": True, "false": False, "null": None}
    _default_get_available_values_method = _values.keys


def CharField(*args, **kwargs):
    warnings.warn(
        "Usage CharField will be deprecated in version 0.2.0, use DjangoCharField import instead",
        DeprecationWarning
    )
    return DjangoCharField(*args, **kwargs)


def NumberField(*args, **kwargs):
    warnings.warn(
        "Usage NumberField will be deprecated in version 0.2.0, use DjangoNumberField import instead",
        DeprecationWarning
    )
    return DjangoNumberField(*args, **kwargs)


def IntegerField(*args, **kwargs):
    warnings.warn(
        "Usage IntegerField will be deprecated in version 0.2.0, use DjangoIntegerField import instead",
        DeprecationWarning
    )
    return DjangoIntegerField(*args, **kwargs)


def FloatField(*args, **kwargs):
    warnings.warn(
        "Usage FloatField will be deprecated in version 0.2.0, use DjangoFloatField import instead",
        DeprecationWarning
    )
    return DjangoFloatField(*args, **kwargs)


def BooleanField(*args, **kwargs):
    warnings.warn(
        "Usage BooleanField will be deprecated in version 0.2.0, use DjangoBooleanField import instead",
        DeprecationWarning
    )
    return DjangoBooleanField(*args, **kwargs)


def NullBooleanField(*args, **kwargs):
    warnings.warn(
        "Usage NullBooleanField will be deprecated in version 0.2.0, use DjangoNullBooleanField import instead",
        DeprecationWarning
    )
    return DjangoNullBooleanField(*args, **kwargs)



from .base import BaseSearchSet
from ..searchset.fields.elastic import ElasticSearchField
from ..parser import LuceneToElasticParserMixin


class ElasticSearchSet(LuceneToElasticParserMixin, BaseSearchSet):
    _field_base_class = ElasticSearchField
    _default_field = ElasticSearchField

    @classmethod
    def filter(cls, search, search_terms):
        query = cls.parse(raw_expression=search_terms)
        return None if query is None else search.query(query)
